/*
 * Stack support for new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/util/elf.h>
#include <l4/util/util.h>

#include <l4/libloader/adjust_stack>

#include <mem/memory_utils.h>

#include <stdio.h>
#include <string.h>

#include "stack.h"



/* Initialise a stack in a memory segment. */

Stack::Stack(Segment &segment)
: _segment(segment)
{
  _element = (l4_umword_t *) (segment.address() + segment.size());

  /* Add a terminator for any additional initial capabilities. */

  l4re_env_cap_entry_t *entry = (l4re_env_cap_entry_t *) _element;

  *(--entry) = l4re_env_cap_entry_t();
  _element = (l4_umword_t *) entry;
}

/* Push any additional initial capabilities. */

void Stack::push_cap_entries(l4re_env_cap_entry_t *entries)
{
  l4re_env_cap_entry_t *entry = (l4re_env_cap_entry_t *) _element;

  while ((entries != NULL) && (entries->cap != L4_INVALID_CAP))
    *(--entry) = *entries;

  _caps = (l4_addr_t) entry;
  _element = (l4_umword_t *) entry;
}

/* Push environment values in reverse order. */

void Stack::push_string(char *s)
{
  char *arg = (char *) _element;
  char *arg_last = arg;

  arg -= round(strlen(s) + 1, sizeof(l4_umword_t));

  memset(arg, 0, arg_last - arg);
  memcpy(arg, s, strlen(s));

  _element = (l4_umword_t *) arg;
}

/* Push environment values in reverse order. */

void Stack::push_env(char *envp[])
{
  _env_top = (char *) _element;
  _env_entries = 0;

  for (; *envp != NULL; envp++, _env_entries++)
    push_string(*envp);
}

/* Push argument values in reverse order. */

void Stack::push_args(int argc, char *argv[])
{
  _arg_top = (char *) _element;

  for (int i = 0; i < argc; i++)
    push_string(argv[i]);
}

/* Loader flags, debugging flags, and the KIP capability index.
   See: generate_l4aux in Remote_app_model */

void Stack::push_l4re_aux()
{
  _aux = (l4re_aux_t *) _element;
  _aux--;
  _element = (l4_umword_t *) _aux;
}

void Stack::push_l4re_env()
{
  _env = (l4re_env_t *) _element;
  _env--;
  _element = (l4_umword_t *) _env;
}

/* Set the common auxiliary information. */

void Stack::set_l4re_aux(l4re_aux_t *aux)
{
  memcpy(_aux, aux, sizeof(l4re_aux_t));

  /* Take the binary name from the first program argument. */

  _aux->binary = (char *) _argv[0];
}

/* Set the common environment, introducing the reference to additional
   capabilities. */

void Stack::set_l4re_env(l4re_env_t *env)
{
  memcpy(_env, env, sizeof(l4re_env_t));
  _env->caps = (l4re_env_cap_entry_t *) (_segment.region_address(_caps));
}

/* Push the auxiliary vector. */

void Stack::push_auxv()
{
  _auxv_end = (char *) _element;

  struct auxv_entry *auxv_base = (struct auxv_entry *) _element;

  *(--auxv_base) = {AT_NULL, 0};

  /* L4Re global environment pointer. */

  *(--auxv_base) = {AT_L4_ENV, _segment.region_address((char *) _env)};

  /* L4Re auxiliary structure pointer. */

  *(--auxv_base) = {AT_L4_AUX, _segment.region_address((char *) _aux)};

  /* Apparently required entries.
     NOTE: The user/group identifiers should be obtained from the broader
           environment. */

  *(--auxv_base) = {AT_PAGESZ, L4_PAGESIZE};
  *(--auxv_base) = {AT_UID, 0};
  *(--auxv_base) = {AT_EUID, 0};
  *(--auxv_base) = {AT_GID, 0};
  *(--auxv_base) = {AT_EGID, 0};

  _element = (l4_umword_t *) auxv_base;
}

/* Fill the stack in reverse with an address, returning the size of the
   value. */

offset_t Stack::write_address(char *arg, char **addr, char *s)
{
  offset_t size = round(strlen(s) + 1, sizeof(l4_umword_t));

  *addr = (char *) _segment.region_address(arg - size);
  return size;
}

/* Populate stack with environment pointers, employing a pointer ordering
   that is the reverse of the value ordering. */

void Stack::push_envp(char *envp[])
{
  /* Write the terminating element. */

  *(--_element) = 0;

  /* Reserve space and fill the stack from the top inwards. */

  char **ep = (char **) (_element - _env_entries);
  char *arg = _env_top;

  for (; *envp != NULL; envp++, ep++)
    arg -= write_address(arg, ep, *envp);

  _element -= _env_entries;
}

/* Populate stack with argument pointers and count, employing a pointer
   ordering that is the reverse of the value ordering. */

void Stack::push_argv(int argc, char *argv[])
{
  /* Write the terminating element. */

  *(--_element) = 0;

  /* Reserve space and fill the stack from the top inwards. */

  _argv = (char **) (_element - argc);
  char *arg = _arg_top;

  for (int i = 0; i < argc; i++)
    arg -= write_address(arg, &_argv[i], argv[i]);

  /* Write the count. */

  _element -= argc;

  *(--_element) = argc;
}

/* Adjust the stack alignment and return the stack address. */

l4_addr_t Stack::align_stack()
{
  char *current = (char *) _element;
  char *adjusted = Ldr::adjust_sp(current, NULL);

  if (adjusted != current)
    memmove(adjusted, (const void *) current, _auxv_end - current);

  _element = (l4_umword_t *) adjusted;

  return _segment.region_address(adjusted);
}

/* Populate the stack with arguments and initial environment. */

void Stack::populate(int argc, char *argv[], char *envp[])
{
  /* Populate stack with environment and argument values. */

  push_env(envp);
  push_args(argc, argv);

  /* Push L4Re flags, environment and auxiliary vector. */

  push_l4re_aux();
  push_l4re_env();
  push_auxv();

  /* Push environment and argument pointers. */

  push_envp(envp);
  push_argv(argc, argv);

  /* Adjust the stack alignment, setting the start address. */

  _start = align_stack();
}

/* Return the stack pointer start address in the mapped program segment. */

l4_addr_t Stack::start_address()
{
  return _start;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
